
/* cairo - a vector graphics library with display and print output
 *
 * Copyright © 2009 Eric Anholt
 * Copyright © 2009 Chris Wilson
 * Copyright © 2005 Red Hat, Inc
 *
 * This library is free software; you can redistribute it and/or
 * modify it either under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation
 * (the "LGPL") or, at your option, under the terms of the Mozilla
 * Public License Version 1.1 (the "MPL"). If you do not alter this
 * notice, a recipient may use your version of this file under either
 * the MPL or the LGPL.
 *
 * You should have received a copy of the LGPL along with this library
 * in the file COPYING-LGPL-2.1; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA
 * You should have received a copy of the MPL along with this library
 * in the file COPYING-MPL-1.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY
 * OF ANY KIND, either express or implied. See the LGPL or the MPL for
 * the specific language governing rights and limitations.
 *
 * The Original Code is the cairo graphics library.
 *
 * The Initial Developer of the Original Code is Red Hat, Inc.
 *
 * Contributor(s):
 *	Carl Worth <cworth@cworth.org>
 *	Chris Wilson <chris@chris-wilson.co.uk>
 */

#include "cairoint.h"

#include "cairo-gl-private.h"

#include "cairo-error-private.h"
#include <dlfcn.h>

typedef struct _cairo_mgl_context
{
    cairo_gl_context_t base;

    MGLContext context;

    MGLSurface dummy_surface;

    MGLContext previous_context;
    MGLSurface previous_surface;

    mgl_make_current_function make_current;
    void *custom_data;
} cairo_mgl_context_t;

typedef struct _cairo_mgl_surface
{
    cairo_gl_surface_t base;

    MGLSurface mgl;
} cairo_mgl_surface_t;

static cairo_bool_t
_context_acquisition_changed_mgl_state(cairo_mgl_context_t *ctx,
                                       MGLSurface current_surface)
{
    if (ctx->make_current)
        return TRUE;

    return ctx->previous_context != ctx->context ||
           ctx->previous_surface != current_surface;
}

static MGLSurface
_mgl_get_current_surface(cairo_mgl_context_t *ctx)
{
    if (ctx->base.current_target == NULL ||
        _cairo_gl_surface_is_texture(ctx->base.current_target))
    {
        return ctx->dummy_surface;
    }

    return ((cairo_mgl_surface_t *)ctx->base.current_target)->mgl;
}

static void
_mgl_query_current_state(cairo_mgl_context_t *ctx)
{
    ctx->previous_surface = mglGetCurrentSurface(MGL_DRAW);
    ctx->previous_context = mglGetCurrentContext();

    /* If any of the values were none, assume they are all none. Not all
       drivers seem well behaved when it comes to using these values across
       multiple threads. */
    if (ctx->previous_surface == MGL_NO_SURFACE ||
        ctx->previous_context == MGL_NO_CONTEXT)
    {
        ctx->previous_surface = MGL_NO_SURFACE;
        ctx->previous_context = MGL_NO_CONTEXT;
    }
}

static void
_mgl_acquire(void *abstract_ctx)
{
    cairo_mgl_context_t *ctx = abstract_ctx;
    MGLSurface current_surface = _mgl_get_current_surface(ctx);

    if (!ctx->make_current)
    {
        _mgl_query_current_state(ctx);
        if (!_context_acquisition_changed_mgl_state(ctx, current_surface))
            return;
    }

    if (ctx->make_current)
    {
        int same_context;
        ctx->make_current(ctx->custom_data, current_surface, ctx->context, &same_context);
    }
    else
        mglMakeCurrent(current_surface, current_surface, ctx->context);
}

static void
_mgl_release(void *abstract_ctx)
{
    cairo_mgl_context_t *ctx = abstract_ctx;
    if (!ctx->base.thread_aware ||
        !_context_acquisition_changed_mgl_state(ctx, _mgl_get_current_surface(ctx)))
    {
        return;
    }

    if (ctx->make_current)
    {
        // nothing
    }
    else
        mglMakeCurrent(MGL_NO_SURFACE, MGL_NO_SURFACE, MGL_NO_CONTEXT);
}

static void
_mgl_make_current(void *abstract_ctx,
                  cairo_gl_surface_t *abstract_surface)
{
    cairo_mgl_context_t *ctx = abstract_ctx;
    cairo_mgl_surface_t *surface = (cairo_mgl_surface_t *)abstract_surface;

    if (ctx->make_current)
    {
        int same_context;
        ctx->make_current(ctx->custom_data, surface->mgl, ctx->context, &same_context);
    }
    else
        mglMakeCurrent(surface->mgl, surface->mgl, ctx->context);
}

static void
_mgl_swap_buffers(void *abstract_ctx,
                  cairo_gl_surface_t *abstract_surface)
{
    cairo_mgl_surface_t *surface = (cairo_mgl_surface_t *)abstract_surface;

    mglSwapBuffers(surface->mgl, MGL_SWAP_WAITGL);
}

static void
_mgl_destroy(void *abstract_ctx)
{
    cairo_mgl_context_t *ctx = abstract_ctx;

    if (ctx->make_current)
    {
        // nothing
    }
    else
        mglMakeCurrent(MGL_NO_SURFACE, MGL_NO_SURFACE, MGL_NO_CONTEXT);
    if (ctx->dummy_surface != MGL_DUMMY_SURFACE && ctx->dummy_surface != MGL_NO_SURFACE)
        mglDestroySurface(ctx->dummy_surface);
}

static cairo_gl_generic_func_t
getProcAddress(const char *name)
{
    return dlsym(RTLD_DEFAULT, name);
}

cairo_device_t *
cairo_mgl_device_create_custom(MGLContext mgl, MGLConfig config, unsigned int vbo_size, mgl_make_current_function make_current, void *custom_data)
{
    cairo_mgl_context_t *ctx;
    cairo_status_t status;

    ctx = calloc(1, sizeof(cairo_mgl_context_t));
    if (unlikely(ctx == NULL))
        return _cairo_gl_context_create_in_error(CAIRO_STATUS_NO_MEMORY);

    ctx->context = mgl;
    ctx->make_current = make_current;
    ctx->custom_data = custom_data;

    ctx->base.acquire = _mgl_acquire;
    ctx->base.release = _mgl_release;
    ctx->base.make_current = _mgl_make_current;
    ctx->base.swap_buffers = _mgl_swap_buffers;
    ctx->base.destroy = _mgl_destroy;

    /* We are about the change the current state of MGL, so we should
     * query the pre-existing surface now instead of later. */
    if (!ctx->make_current)
        _mgl_query_current_state(ctx);


    ctx->dummy_surface = MGL_DUMMY_SURFACE; // MGL doesn't support surfaceless rendering, although you can use MGL_DUMMY_SURFACE instead.

    if (ctx->make_current)
    {
        int same_context;
        ctx->make_current(ctx->custom_data, ctx->dummy_surface, mgl, &same_context);
    }
    else
        mglMakeCurrent(ctx->dummy_surface, ctx->dummy_surface, mgl);

    status = _cairo_gl_dispatch_init(&ctx->base.dispatch, getProcAddress);
    if (unlikely(status))
    {
        free(ctx);
        return _cairo_gl_context_create_in_error(status);
    }

    status = _cairo_gl_context_init(&ctx->base);
    if (unlikely(status))
    {
        free(ctx);
        return _cairo_gl_context_create_in_error(status);
    }

    /* Tune the default VBO size to reduce overhead on embedded devices.
     * This smaller size means that flushing needs to be done more often,
     * but it is less demanding of scarce memory on embedded devices.
     */
    ctx->base.vbo_size = 64 * 1024;

    if (vbo_size > 0)
        ctx->base.vbo_size = vbo_size;

    if (ctx->make_current)
    {
        int same_context;
        ctx->make_current(ctx->custom_data, MGL_NO_SURFACE, MGL_NO_CONTEXT, &same_context);
    }
    else
        mglMakeCurrent(MGL_NO_SURFACE, MGL_NO_SURFACE, MGL_NO_CONTEXT);

    return &ctx->base.base;
}

cairo_device_t *
cairo_mgl_device_create(MGLContext mgl)
{

    //this function is not supported on mgl
    return _cairo_gl_context_create_in_error(CAIRO_STATUS_DEVICE_ERROR);
}

cairo_surface_t *
cairo_gl_surface_create_for_mgl(cairo_device_t *device,
                                MGLSurface mgl,
                                int width,
                                int height)
{
    cairo_mgl_surface_t *surface;

    if (unlikely(device->status))
        return _cairo_surface_create_in_error(device->status);

    if (device->backend->type != CAIRO_DEVICE_TYPE_GL)
        return _cairo_surface_create_in_error(_cairo_error(CAIRO_STATUS_SURFACE_TYPE_MISMATCH));

    if (width <= 0 || height <= 0)
        return _cairo_surface_create_in_error(_cairo_error(CAIRO_STATUS_INVALID_SIZE));

    surface = calloc(1, sizeof(cairo_mgl_surface_t));
    if (unlikely(surface == NULL))
        return _cairo_surface_create_in_error(_cairo_error(CAIRO_STATUS_NO_MEMORY));

    _cairo_gl_surface_init(device, &surface->base,
                           CAIRO_CONTENT_COLOR_ALPHA, width, height);
    surface->mgl = mgl;

    return &surface->base.base;
}

static cairo_bool_t is_mgl_device(cairo_device_t *device)
{
    return (device->backend != NULL &&
            device->backend->type == CAIRO_DEVICE_TYPE_GL);
}

static cairo_mgl_context_t *to_mgl_context(cairo_device_t *device)
{
    return (cairo_mgl_context_t *)device;
}

cairo_public MGLContext
cairo_mgl_device_get_context(cairo_device_t *device)
{
    if (!is_mgl_device(device))
    {
        _cairo_error_throw(CAIRO_STATUS_DEVICE_TYPE_MISMATCH);
        return MGL_NO_CONTEXT;
    }

    return to_mgl_context(device)->context;
}
